"""Base class providing common storage helpers (threading, MIME guess, futures)."""

import mimetypes
from concurrent.futures import ThreadPoolExecutor, wait

from datacosmos.datacosmos_client import DatacosmosClient


class StorageBase:
    """Base class providing common storage helpers (threading, MIME guess, futures)."""

    def __init__(self, client: DatacosmosClient):
        """Base class providing common storage helpers (threading, MIME guess, futures)."""
        self.client = client
        self.base_url = client.config.datacosmos_cloud_storage.as_domain_url()

    def _guess_mime(self, src: str) -> str:
        mime, _ = mimetypes.guess_type(src)
        return mime or "application/octet-stream"

    def _run_in_threads(self, fn, fn_args, max_workers: int, timeout: float):
        """Run the callable `fn(*args)` over the iterable of jobs in parallel threads.

        `jobs` should be a list of tuples, each tuple unpacked as fn(*args).
        """
        futures = []
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for args in fn_args:
                futures.append(executor.submit(fn, *args))
        done, not_done = wait(futures, timeout=timeout)
        errors = []
        for future in done:
            try:
                future.result()
            except Exception as e:
                errors.append(e)
        for future in not_done:
            future.cancel()
        if errors:
            raise errors[0]



"""Unified interface for STAC API, combining Item & Collection operations."""

from datacosmos.stac.collection.collection_client import CollectionClient
from datacosmos.stac.item.item_client import ItemClient
from datacosmos.stac.storage.storage_client import StorageClient


class STACClient(ItemClient, CollectionClient, StorageClient):
    """Unified interface for STAC API, combining Item & Collection operations."""

    def __init__(self, client):
        """Initialize the STACClient with a DatacosmosClient."""
        ItemClient.__init__(self, client)
        CollectionClient.__init__(self, client)
        StorageClient.__init__(self, client)



"""Client to interact with the Datacosmos API with authentication and request handling."""

from datetime import datetime, timedelta, timezone
from typing import Any, Optional

import requests
from oauthlib.oauth2 import BackendApplicationClient
from requests.exceptions import ConnectionError, HTTPError, RequestException, Timeout
from requests_oauthlib import OAuth2Session

from datacosmos.config.config import Config
from datacosmos.exceptions.datacosmos_exception import DatacosmosException


class DatacosmosClient:
    """Client to interact with the Datacosmos API with authentication and request handling."""

    def __init__(
        self,
        config: Optional[Config | Any] = None,
        http_session: Optional[requests.Session | OAuth2Session] = None,
    ):
        """Initialize the DatacosmosClient.

        Args:
            config (Optional[Config]): Configuration object (only needed when SDK creates its own session).
            http_session (Optional[requests.Session]): Pre-authenticated session.
        """
        if http_session is not None:
            self._http_client = http_session
            self._owns_session = False
            if isinstance(http_session, OAuth2Session):
                token_data = http_session.token
            elif isinstance(http_session, requests.Session):
                auth_header = http_session.headers.get("Authorization", "")
                if not auth_header.startswith("Bearer "):
                    raise DatacosmosException(
                        "Injected requests.Session must include a 'Bearer' token in its headers"
                    )
                token_data = {"access_token": auth_header.split(" ", 1)[1]}
            else:
                raise DatacosmosException(
                    f"Unsupported session type: {type(http_session)}"
                )
            try:
                self.token = token_data.get("access_token")
                self.token_expiry = token_data.get("expires_at") or token_data.get(
                    "expires_in"
                )
            except Exception:
                raise DatacosmosException(
                    "Failed to extract token from injected session"
                )

            self.config = config
        else:
            if config:
                self.config = config
            else:
                try:
                    self.config = Config.from_yaml()
                except ValueError:
                    self.config = Config.from_env()

            self._owns_session = True
            self.token = None
            self.token_expiry = None
            self._http_client = self._authenticate_and_initialize_client()

    def _authenticate_and_initialize_client(self) -> requests.Session:
        """Authenticate and initialize the HTTP client with a valid token."""
        try:
            client = BackendApplicationClient(
                client_id=self.config.authentication.client_id
            )
            oauth_session = OAuth2Session(client=client)

            token_response = oauth_session.fetch_token(
                token_url=self.config.authentication.token_url,
                client_id=self.config.authentication.client_id,
                client_secret=self.config.authentication.client_secret,
                audience=self.config.authentication.audience,
            )

            self.token = token_response["access_token"]
            self.token_expiry = datetime.now(timezone.utc) + timedelta(
                seconds=token_response.get("expires_in", 3600)
            )

            http_client = requests.Session()
            http_client.headers.update({"Authorization": f"Bearer {self.token}"})
            return http_client
        except (HTTPError, ConnectionError, Timeout) as e:
            raise DatacosmosException(f"Authentication failed: {str(e)}") from e
        except RequestException as e:
            raise DatacosmosException(
                f"Unexpected request failure during authentication: {str(e)}"
            ) from e

    def _refresh_token_if_needed(self):
        """Refresh the token if it has expired (only if SDK created it)."""
        if self._owns_session and (
            not self.token or self.token_expiry <= datetime.now(timezone.utc)
        ):
            self._http_client = self._authenticate_and_initialize_client()

    def request(
        self, method: str, url: str, *args: Any, **kwargs: Any
    ) -> requests.Response:
        """Send an HTTP request using the authenticated session."""
        self._refresh_token_if_needed()
        try:
            response = self._http_client.request(method, url, *args, **kwargs)
            response.raise_for_status()
            return response
        except HTTPError as e:
            raise DatacosmosException(
                f"HTTP error during {method.upper()} request to {url}",
                response=e.response,
            ) from e
        except ConnectionError as e:
            raise DatacosmosException(
                f"Connection error during {method.upper()} request to {url}: {str(e)}"
            ) from e
        except Timeout as e:
            raise DatacosmosException(
                f"Request timeout during {method.upper()} request to {url}: {str(e)}"
            ) from e
        except RequestException as e:
            raise DatacosmosException(
                f"Unexpected request failure during {method.upper()} request to {url}: {str(e)}"
            ) from e

    def get(self, url: str, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a GET request using the authenticated session."""
        return self.request("GET", url, *args, **kwargs)

    def post(self, url: str, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a POST request using the authenticated session."""
        return self.request("POST", url, *args, **kwargs)

    def put(self, url: str, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a PUT request using the authenticated session."""
        return self.request("PUT", url, *args, **kwargs)

    def patch(self, url: str, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a PATCH request using the authenticated session."""
        return self.request("PATCH", url, *args, **kwargs)

    def delete(self, url: str, *args: Any, **kwargs: Any) -> requests.Response:
        """Send a DELETE request using the authenticated session."""
        return self.request("DELETE", url, *args, **kwargs)



"""Represents a structured update model for STAC collections.

Allows partial updates where only the provided fields are modified.
"""

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field
from pystac import Extent, Link, Provider, Summaries


class CollectionUpdate(BaseModel):
    """Represents a structured update model for STAC collections.

    Allows partial updates where only the provided fields are modified.
    """

    model_config = {"arbitrary_types_allowed": True}

    title: Optional[str] = Field(None, description="Title of the STAC collection.")
    description: Optional[str] = Field(
        None, description="Description of the collection."
    )
    keywords: Optional[List[str]] = Field(
        None, description="List of keywords associated with the collection."
    )
    license: Optional[str] = Field(None, description="Collection license information.")
    providers: Optional[List[Provider]] = Field(
        None, description="List of data providers."
    )
    extent: Optional[Extent] = Field(
        None, description="Spatial and temporal extent of the collection."
    )
    summaries: Optional[Summaries] = Field(
        None, description="Summaries for the collection."
    )
    links: Optional[List[Link]] = Field(
        None, description="List of links associated with the collection."
    )

    def to_dict(self) -> Dict[str, Any]:
        """Convert the model into a dictionary, excluding `None` values.

        Returns:
            Dict[str, Any]: Dictionary representation of the update payload.
        """
        return self.model_dump(by_alias=True, exclude_none=True)



"""Structured API error message for Datacosmos."""

from pydantic import BaseModel


class DatacosmosError(BaseModel):
    """Structured API error message for Datacosmos."""

    message: str
    field: str | None = None
    type: str | None = None
    source: str | None = None
    trace_id: str | None = None

    def human_readable(self) -> str:
        """Formats the error message into a readable format."""
        msg = self.message
        if self.type:
            msg += f" (type: {self.type})"
        if self.field:
            msg += f" (field: {self.field})"
        if self.source:
            msg += f" (source: {self.source})"
        if self.trace_id:
            msg += f" (trace_id: {self.trace_id})"
        return msg



"""Base exception class for all Datacosmos SDK exceptions."""

from typing import Optional

from requests import Response
from requests.exceptions import RequestException


class DatacosmosException(RequestException):
    """Base exception class for all Datacosmos SDK exceptions."""

    def __init__(self, message: str, response: Optional[Response] = None):
        """Initialize DatacosmosException.

        Args:
            message (str): The error message.
            response (Optional[Response]): The HTTP response object, if available.
        """
        self.response = response
        self.status_code = response.status_code if response else None
        self.details = response.text if response else None
        full_message = (
            f"{message} (Status: {self.status_code}, Details: {self.details})"
            if response
            else message
        )
        super().__init__(full_message)



"""Configuration module for the Datacosmos SDK.

Handles configuration management using Pydantic and Pydantic Settings.
It loads default values, allows overrides via YAML configuration files,
and supports environment variable-based overrides.
"""

import os
from typing import ClassVar, Optional

import yaml
from pydantic import field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from datacosmos.config.models.authentication_config import AuthenticationConfig
from datacosmos.config.models.m2m_authentication_config import M2MAuthenticationConfig
from datacosmos.config.models.url import URL


class Config(BaseSettings):
    """Centralized configuration for the Datacosmos SDK."""

    model_config = SettingsConfigDict(
        env_nested_delimiter="__",
        nested_model_default_partial_update=True,
        extra="allow",
    )

    authentication: Optional[AuthenticationConfig] = None
    stac: Optional[URL] = None
    datacosmos_cloud_storage: Optional[URL] = None
    datacosmos_public_cloud_storage: Optional[URL] = None

    DEFAULT_AUTH_TYPE: ClassVar[str] = "m2m"
    DEFAULT_AUTH_TOKEN_URL: ClassVar[str] = "https://login.open-cosmos.com/oauth/token"
    DEFAULT_AUTH_AUDIENCE: ClassVar[str] = "https://beeapp.open-cosmos.com"

    @classmethod
    def from_yaml(cls, file_path: str = "config/config.yaml") -> "Config":
        """Load configuration from a YAML file and override defaults.

        Args:
            file_path (str): The path to the YAML configuration file.

        Returns:
            Config: An instance of the Config class with loaded settings.
        """
        config_data: dict = {}
        if os.path.exists(file_path):
            with open(file_path, "r") as f:
                yaml_data = yaml.safe_load(f) or {}
                # Remove empty values from YAML to avoid overwriting with `None`
                config_data = {
                    key: value
                    for key, value in yaml_data.items()
                    if value not in [None, ""]
                }

        return cls(**config_data)

    @classmethod
    def from_env(cls) -> "Config":
        """Load configuration from environment variables.

        Returns:
            Config: An instance of the Config class with settings loaded from environment variables.
        """
        authentication_config = M2MAuthenticationConfig(
            type=os.getenv("OC_AUTH_TYPE", cls.DEFAULT_AUTH_TYPE),
            client_id=os.getenv("OC_AUTH_CLIENT_ID"),
            client_secret=os.getenv("OC_AUTH_CLIENT_SECRET"),
            token_url=os.getenv("OC_AUTH_TOKEN_URL", cls.DEFAULT_AUTH_TOKEN_URL),
            audience=os.getenv("OC_AUTH_AUDIENCE", cls.DEFAULT_AUTH_AUDIENCE),
        )

        stac_config = URL(
            protocol=os.getenv("OC_STAC_PROTOCOL", "https"),
            host=os.getenv("OC_STAC_HOST", "app.open-cosmos.com"),
            port=int(os.getenv("OC_STAC_PORT", "443")),
            path=os.getenv("OC_STAC_PATH", "/api/data/v0/stac"),
        )

        datacosmos_cloud_storage_config = URL(
            protocol=os.getenv("DC_CLOUD_STORAGE_PROTOCOL", "https"),
            host=os.getenv("DC_CLOUD_STORAGE_HOST", "app.open-cosmos.com"),
            port=int(os.getenv("DC_CLOUD_STORAGE_PORT", "443")),
            path=os.getenv("DC_CLOUD_STORAGE_PATH", "/api/data/v0/storage"),
        )

        datacosmos_public_cloud_storage_config = URL(
            protocol=os.getenv("DC_PUBLIC_CLOUD_STORAGE_PROTOCOL", "https"),
            host=os.getenv("DC_PUBLIC_CLOUD_STORAGE_HOST", "app.open-cosmos.com"),
            port=int(os.getenv("DC_PUBLIC_CLOUD_STORAGE_PORT", "443")),
            path=os.getenv("DC_PUBLIC_CLOUD_STORAGE_PATH", "/api/data/v0/storage"),
        )

        return cls(
            authentication=authentication_config,
            stac=stac_config,
            datacosmos_cloud_storage=datacosmos_cloud_storage_config,
            datacosmos_public_cloud_storage=datacosmos_public_cloud_storage_config,
        )

    @field_validator("authentication", mode="after")
    @classmethod
    def validate_authentication(
        cls, auth: Optional[M2MAuthenticationConfig]
    ) -> M2MAuthenticationConfig:
        """Ensure authentication is provided and apply defaults.

        Args:
            auth (Optional[M2MAuthenticationConfig]): The authentication config.

        Returns:
            M2MAuthenticationConfig: The validated authentication configuration.

        Raises:
            ValueError: If authentication is missing or required fields are not set.
        """
        if auth is None:
            auth = cls.apply_auth_defaults(M2MAuthenticationConfig())
        else:
            auth = cls.apply_auth_defaults(auth)

        cls.check_required_auth_fields(auth)
        return auth

    @staticmethod
    def apply_auth_defaults(auth: M2MAuthenticationConfig) -> M2MAuthenticationConfig:
        """Apply default authentication values if they are missing."""
        auth.type = auth.type or Config.DEFAULT_AUTH_TYPE
        auth.token_url = auth.token_url or Config.DEFAULT_AUTH_TOKEN_URL
        auth.audience = auth.audience or Config.DEFAULT_AUTH_AUDIENCE
        return auth

    @classmethod
    def parse_auth_config(cls, auth_data: dict) -> M2MAuthenticationConfig:
        """Parse authentication config from a dictionary."""
        return M2MAuthenticationConfig(
            type=auth_data.get("type", cls.DEFAULT_AUTH_TYPE),
            token_url=auth_data.get("token_url", cls.DEFAULT_AUTH_TOKEN_URL),
            audience=auth_data.get("audience", cls.DEFAULT_AUTH_AUDIENCE),
            client_id=auth_data.get("client_id"),
            client_secret=auth_data.get("client_secret"),
        )

    @staticmethod
    def check_required_auth_fields(auth: M2MAuthenticationConfig):
        """Ensure required fields (client_id, client_secret) are provided."""
        missing_fields = [
            field
            for field in ("client_id", "client_secret")
            if not getattr(auth, field)
        ]
        if missing_fields:
            raise ValueError(
                f"Missing required authentication fields: {', '.join(missing_fields)}"
            )

    @field_validator("stac", mode="before")
    @classmethod
    def validate_stac(cls, stac_config: Optional[URL]) -> URL:
        """Ensure STAC configuration has a default if not explicitly set.

        Args:
            stac_config (Optional[URL]): The STAC config to validate.

        Returns:
            URL: The validated STAC configuration.
        """
        if stac_config is None:
            return URL(
                protocol="https",
                host="app.open-cosmos.com",
                port=443,
                path="/api/data/v0/stac",
            )
        return stac_config

    @field_validator("datacosmos_cloud_storage", mode="before")
    @classmethod
    def validate_datacosmos_cloud_storage(
        cls, datacosmos_cloud_storage_config: Optional[URL]
    ) -> URL:
        """Ensure datacosmos cloud storage configuration has a default if not explicitly set.

        Args:
            datacosmos_cloud_storage_config (Optional[URL]): The datacosmos cloud storage config to validate.

        Returns:
            URL: The validated datacosmos cloud storage configuration.
        """
        if datacosmos_cloud_storage_config is None:
            return URL(
                protocol="https",
                host="app.open-cosmos.com",
                port=443,
                path="/api/data/v0/storage",
            )
        return datacosmos_cloud_storage_config

    @field_validator("datacosmos_public_cloud_storage", mode="before")
    @classmethod
    def validate_datacosmos_public_cloud_storage(
        cls, datacosmos_public_cloud_storage_config: Optional[URL]
    ) -> URL:
        """Ensure datacosmos cloud storage configuration has a default if not explicitly set.

        Args:
            datacosmos_public_cloud_storage_config (Optional[URL]): The datacosmos public cloud storage config to validate.

        Returns:
            URL: The validated datacosmos public cloud storage configuration.
        """
        if datacosmos_public_cloud_storage_config is None:
            return URL(
                protocol="https",
                host="app.open-cosmos.com",
                port=443,
                path="/api/data/v0/storage",
            )
        return datacosmos_public_cloud_storage_config



"""STAC package for interacting with items from the STAC API, providing query and fetch functionalities.

It enables interaction with items from the STAC using an authenticated Datacosmos client.
"""



"""Structured response for Datacosmos handling multiple API errors."""

from pydantic import BaseModel

from datacosmos.utils.http_response.models.datacosmos_error import DatacosmosError


class DatacosmosResponse(BaseModel):
    """Structured response for Datacosmos handling multiple API errors."""

    errors: list[DatacosmosError]



"""Model representing a partial update for a STAC item."""

from typing import Any, Optional

from pydantic import BaseModel, Field, model_validator
from pystac import Asset, Link


class ItemUpdate(BaseModel):
    """Model representing a partial update for a STAC item."""

    model_config = {"arbitrary_types_allowed": True}

    stac_extensions: Optional[list[str]] = None
    geometry: Optional[dict[str, Any]] = None
    bbox: Optional[list[float]] = Field(
        None, min_items=4, max_items=4
    )  # Must be [minX, minY, maxX, maxY]
    properties: Optional[dict[str, Any]] = None
    assets: Optional[dict[str, Asset]] = None
    links: Optional[list[Link]] = None

    def set_geometry(self, geom_type: str, coordinates: list[Any]) -> None:
        """Set the geometry manually without using shapely.

        Args:
            geom_type (str): The type of geometry (e.g., 'Point', 'Polygon').
            coordinates (list[Any]): The coordinates defining the geometry.
        """
        self.geometry = {"type": geom_type, "coordinates": coordinates}

    @staticmethod
    def has_valid_datetime(properties: dict[str, Any]) -> bool:
        """Check if 'datetime' is present and not None."""
        return properties.get("datetime") is not None

    @staticmethod
    def has_valid_datetime_range(properties: dict[str, Any]) -> bool:
        """Check if both 'start_datetime' and 'end_datetime' are present and not None."""
        return all(
            properties.get(key) is not None
            for key in ["start_datetime", "end_datetime"]
        )

    @model_validator(mode="before")
    def validate_datetime_fields(cls, values):
        """Ensure at least one of 'datetime' or 'start_datetime'/'end_datetime' exists."""
        properties = values.get("properties", {})

        if not cls.has_valid_datetime(properties) and not cls.has_valid_datetime_range(
            properties
        ):
            raise ValueError(
                "Either 'datetime' or both 'start_datetime' and 'end_datetime' must be provided."
            )

        return values



"""Model representing a datacosmos item asset."""

from pydantic import BaseModel, Field

from datacosmos.stac.item.models.eo_band import EoBand
from datacosmos.stac.item.models.raster_band import RasterBand


class Asset(BaseModel):
    """Model representing a datacosmos item asset."""

    href: str
    title: str
    description: str
    type: str
    roles: list[str] | None
    eo_bands: list[EoBand] | None = Field(default=None, alias="eo:bands")
    raster_bands: list[RasterBand] | None = Field(default=None, alias="raster:bands")

    class Config:
        """Pydantic configuration."""

        populate_by_name = True



"""Handles uploading files to Datacosmos storage and registering STAC items."""

from pathlib import Path

from pydantic import TypeAdapter

from datacosmos.datacosmos_client import DatacosmosClient
from datacosmos.stac.item.item_client import ItemClient
from datacosmos.stac.item.models.asset import Asset
from datacosmos.stac.item.models.datacosmos_item import DatacosmosItem
from datacosmos.stac.storage.dataclasses.upload_path import UploadPath
from datacosmos.stac.storage.storage_base import StorageBase


class Uploader(StorageBase):
    """Handles uploading files to Datacosmos storage and registering STAC items."""

    def __init__(self, client: DatacosmosClient):
        """Handles uploading files to Datacosmos storage and registering STAC items."""
        super().__init__(client)
        self.item_client = ItemClient(client)

    def upload_item(
        self,
        item: DatacosmosItem,
        assets_path: str | None = None,
        included_assets: list[str] | bool = True,
        max_workers: int = 4,
        time_out: float = 60 * 60 * 1,
    ) -> DatacosmosItem:
        """Upload a STAC item and its assets to Datacosmos."""
        if not assets_path and not isinstance(item, str):
            raise ValueError(
                "assets_path must be provided if item is not the path to an item file."
            )

        if isinstance(item, str):
            item_filename = item
            item = self._load_item(item_filename)
            if not assets_path:
                assets_path = str(Path(item_filename).parent)

        assets_path = assets_path or str(Path.cwd())

        upload_assets = (
            included_assets
            if isinstance(included_assets, list)
            else item.assets.keys()
            if included_assets is True
            else []
        )

        jobs = [(item, asset_key, assets_path) for asset_key in upload_assets]

        self._run_in_threads(self._upload_asset, jobs, max_workers, time_out)

        self.item_client.add_item(item)

        return item

    def upload_from_file(
        self, src: str, dst: str, mime_type: str | None = None
    ) -> None:
        """Uploads a single file to the specified destination path."""
        url = self.base_url.with_suffix(dst)
        mime = mime_type or self._guess_mime(src)
        headers = {"Content-Type": mime}
        with open(src, "rb") as f:
            response = self.client.put(url, data=f, headers=headers)
        response.raise_for_status()

    @staticmethod
    def _load_item(item_json_file_path: str) -> DatacosmosItem:
        with open(item_json_file_path, "rb") as file:
            data = file.read().decode("utf-8")
        return TypeAdapter(DatacosmosItem).validate_json(data)

    def _upload_asset(
        self, item: DatacosmosItem, asset_key: str, assets_path: str
    ) -> None:
        asset = item.assets[asset_key]
        upload_path = UploadPath.from_item_path(item, "", Path(asset.href).name)
        local_src = Path(assets_path) / asset.href
        if local_src.exists():
            src = str(local_src)
            asset.href = f"file:///{upload_path}"
        else:
            src = str(Path(assets_path) / Path(asset.href).name)
        self._update_asset_href(asset)
        self.upload_from_file(src, str(upload_path), mime_type=asset.type)

    def _update_asset_href(self, asset: Asset) -> None:
        try:
            url = self.client.config.datacosmos_public_cloud_storage.as_domain_url()
            new_href = url.with_base(asset.href)  # type: ignore
            asset.href = str(new_href)
        except ValueError:
            pass



"""Model representing a raster band."""

from pydantic import BaseModel, Field


class RasterBand(BaseModel):
    """Model representing a raster band."""

    gain: float = Field(alias="scale")
    bias: float = Field(alias="offset")
    nodata: int
    unit: str

    class Config:
        """Pydantic configuration."""

        populate_by_name = True



"""Model representing a datacosmos item."""

from datetime import datetime

from pydantic import BaseModel

from datacosmos.stac.enums.processing_level import ProcessingLevel
from datacosmos.stac.item.models.asset import Asset


class DatacosmosItem(BaseModel):
    """Model representing a datacosmos item."""

    id: str
    type: str
    stac_version: str
    stac_extensions: list | None
    geometry: dict
    properties: dict
    links: list
    assets: dict[str, Asset]
    collection: str
    bbox: tuple[float, float, float, float]

    def get_property(self, key: str) -> str | None:
        """Get a property value from the Datacosmos item."""
        return self.properties.get(key)

    def get_asset(self, key: str) -> Asset | None:
        """Get an asset from the Datacosmos item."""
        return self.assets.get(key)

    @property
    def datetime(self) -> datetime:
        """Get the datetime of the Datacosmos item."""
        return datetime.strptime(self.properties["datetime"], "%Y-%m-%dT%H:%M:%SZ")

    @property
    def level(self) -> ProcessingLevel:
        """Get the processing level of the Datacosmos item."""
        return ProcessingLevel(self.properties["processing:level"].lower())

    @property
    def sat_int_designator(self) -> str:
        """Get the satellite international designator of the Datacosmos item."""
        property = self.get_property("sat:platform_international_designator")
        if property is None:
            raise ValueError(
                "sat:platform_international_designator is missing in STAC item"
            )
        return property

    def to_dict(self) -> dict:
        """Converts the DatacosmosItem instance to a dictionary."""
        return self.model_dump()



"""Validates an API response and raises a DatacosmosException if an error occurs."""

from pydantic import ValidationError
from requests import Response

from datacosmos.exceptions.datacosmos_exception import DatacosmosException
from datacosmos.utils.http_response.models.datacosmos_response import DatacosmosResponse


def check_api_response(response: Response) -> None:
    """Validates an API response and raises a DatacosmosException if an error occurs.

    Args:
        resp (requests.Response): The response object.

    Raises:
        DatacosmosException: If the response status code indicates an error.
    """
    if 200 <= response.status_code < 400:
        return

    try:
        response = DatacosmosResponse.model_validate_json(response.text)
        msg = response.errors[0].human_readable()
        if len(response.errors) > 1:
            msg = "\n  * " + "\n  * ".join(
                error.human_readable() for error in response.errors
            )
        raise DatacosmosException(msg, response=response)

    except ValidationError:
        raise DatacosmosException(
            f"HTTP {response.status_code}: {response.text}", response=response
        )



"""Model representing an EO band."""

from typing import Optional

from pydantic import BaseModel


class EoBand(BaseModel):
    """Model representing an EO band."""

    name: str
    common_name: str
    center_wavelength: float
    full_width_half_max: float
    solar_illumination: Optional[float] = None



"""Test suite for basic functionality and CI setup."""


class TestPass:
    """A simple test class to validate the CI pipeline setup."""

    def test_pass(self):
        """A passing test to ensure the CI pipeline is functional."""
        assert True



"""Generic StorageClient for all storage operations (upload, download, etc.)."""

from datacosmos.datacosmos_client import DatacosmosClient
from datacosmos.stac.item.models.datacosmos_item import DatacosmosItem
from datacosmos.stac.storage.uploader import Uploader


class StorageClient:
    """Generic StorageClient for all storage operations (upload, download, etc.)."""

    def __init__(self, client: DatacosmosClient):
        """Generic StorageClient for all storage operations (upload, download, etc.)."""
        self.client = client
        self.uploader = Uploader(client)

    def upload_item(
        self,
        item: DatacosmosItem,
        assets_path: str | None = None,
        included_assets: list[str] | bool = True,
        max_workers: int = 4,
        time_out: float = 60 * 60 * 1,
    ) -> DatacosmosItem:
        """Proxy to Uploader.upload_item, without needing to pass client each call."""
        return self.uploader.upload_item(
            item=item,
            assets_path=assets_path,
            included_assets=included_assets,
            max_workers=max_workers,
            time_out=time_out,
        )



"""STAC package for interacting with the STAC API, providing query and fetch functionalities.

It enables interaction with STAC (SpatioTemporal Asset Catalog) services
using an authenticated Datacosmos client.
"""



"""Level enum class."""

from enum import Enum


class ProcessingLevel(Enum):
    """Enum class for the processing levels of the data."""

    RAW = "RAW"
    L0 = "l0"
    L1A = "l1A"
    L2A = "l2A"
    L1B = "l1B"
    L1C = "l1C"
    L1D = "l1D"
    L3 = "l3"



"""No authentication configuration.

This is used when running scripts in tests.
"""

from typing import Literal

from pydantic import BaseModel


class NoAuthenticationConfig(BaseModel):
    """No authentication configuration.

    This is used when running scripts in tests.
    """

    type: Literal["none"]



"""Query parameters for catalog search."""

from datetime import datetime, timedelta
from typing import Any, List, Optional

from pydantic import BaseModel, Field, field_validator, model_validator

from datacosmos.stac.constants.satellite_name_mapping import SATELLITE_NAME_MAPPING
from datacosmos.stac.enums.processing_level import ProcessingLevel
from datacosmos.stac.enums.product_type import ProductType
from datacosmos.stac.enums.season import Season


class CatalogSearchParameters(BaseModel):
    """Query parameters for catalog search."""

    start_date: Optional[str] = None
    end_date: Optional[str] = None
    seasons: Optional[List[Season]] = None
    satellite: Optional[List[str]] = None
    product_type: Optional[List[ProductType]] = None
    processing_level: Optional[List[ProcessingLevel]] = None
    collections: Optional[list[str]] = Field(
        None,
        description="Array of collection IDs to filter by.",
        example=["collection1", "collection2"],
    )

    # --- Field Validators ---

    @field_validator("seasons", mode="before")
    @classmethod
    def parse_seasons(cls, value):
        """Parses seasons values into a list of Season object."""
        if value is None:
            return None
        return [Season(v) if not isinstance(v, Season) else v for v in value]

    @field_validator("product_type", mode="before")
    @classmethod
    def parse_product_types(cls, value):
        """Parses product types values into a list of ProductType object."""
        if value is None:
            return None
        return [ProductType(v) if not isinstance(v, ProductType) else v for v in value]

    @field_validator("processing_level", mode="before")
    @classmethod
    def parse_processing_levels(cls, value):
        """Parses processing levels values into a list of ProcessingLevel object."""
        if value is None:
            return None
        return [
            ProcessingLevel(v) if not isinstance(v, ProcessingLevel) else v
            for v in value
        ]

    @field_validator("start_date", mode="before")
    @classmethod
    def parse_start_date(cls, value: Any) -> Optional[str]:
        """Validations on start_date."""
        if value is None:
            return None
        try:
            dt = datetime.strptime(value, "%m/%d/%Y")
        except Exception as e:
            raise ValueError(
                "Invalid start_date format. Use mm/dd/yyyy (e.g., 05/15/2024)"
            ) from e
        if dt < datetime(2015, 5, 15):
            raise ValueError("Date must be 5/15/2015 or later.")
        return dt.isoformat() + "Z"

    @field_validator("end_date", mode="before")
    @classmethod
    def parse_end_date(cls, value: Any) -> Optional[str]:
        """Validations on end_date."""
        if value is None:
            return None
        try:
            dt = datetime.strptime(value, "%m/%d/%Y")
        except ValueError:
            raise ValueError(
                "Invalid end_date format. Use mm/dd/yyyy (e.g., 05/15/2024)"
            )

        if dt < datetime(2015, 5, 15):
            raise ValueError("Date must be 5/15/2015 or later.")
        dt = dt + timedelta(days=1) - timedelta(milliseconds=1)
        return dt.isoformat() + "Z"

    # --- Model Validator ---

    @model_validator(mode="after")
    def validate_date_range(self) -> "CatalogSearchParameters":
        """Checks if end_date is after the start_date."""
        if self.start_date and self.end_date:
            start_dt = datetime.fromisoformat(self.start_date.rstrip("Z"))
            end_dt = datetime.fromisoformat(self.end_date.rstrip("Z"))
            if start_dt > end_dt:
                raise ValueError("end_date cannot be before start_date.")
        return self

    # --- Query Mapper ---

    def to_query(self) -> dict:
        """Map user-friendly input to STAC query structure."""
        query = {}

        if self.start_date or self.end_date:
            query["datetime"] = {"gte": self.start_date, "lte": self.end_date}

        if self.seasons:
            query["opencosmos:season"] = {
                "in": [seasons.value for seasons in self.seasons]
            }

        if self.product_type:
            query["opencosmos:product_type"] = {
                "in": [product_type.value for product_type in self.product_type]
            }

        if self.processing_level:
            query["processing:level"] = {
                "in": [
                    processing_level.value for processing_level in self.processing_level
                ]
            }

        if self.satellite:
            cospars = [
                SATELLITE_NAME_MAPPING[ui]
                for ui in self.satellite
                if ui in SATELLITE_NAME_MAPPING
            ]
            query["sat:platform_international_designator"] = {"in": cospars}

        return query



"""Dataclass for retrieving the upload path of a file."""

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import structlog

from datacosmos.stac.enums.processing_level import ProcessingLevel
from datacosmos.stac.item.models.datacosmos_item import DatacosmosItem

logger = structlog.get_logger()


@dataclass
class UploadPath:
    """Dataclass for retrieving the upload path of a file."""

    mission: str
    level: ProcessingLevel
    day: int
    month: int
    year: int
    id: str
    path: str

    def __str__(self):
        """Return a human-readable string representation of the Path."""
        path = f"full/{self.mission.lower()}/{self.level.value.lower()}/{self.year:02}/{self.month:02}/{self.day:02}/{self.id}/{self.path}"
        return path.removesuffix("/")

    @classmethod
    def from_item_path(
        cls, item: DatacosmosItem, mission: str, item_path: str
    ) -> "Path":
        """Create a Path instance from a DatacosmosItem and a path."""
        dt = datetime.strptime(item.properties["datetime"], "%Y-%m-%dT%H:%M:%SZ")
        path = UploadPath(
            mission=mission,
            level=ProcessingLevel(item.properties["processing:level"]),
            day=dt.day,
            month=dt.month,
            year=dt.year,
            id=item.id,
            path=item_path,
        )
        return cls(**path.__dict__)

    @classmethod
    def from_path(cls, path: str) -> "Path":
        """Create a Path instance from a string path."""
        parts = path.split("/")
        if len(parts) < 7:
            raise ValueError(f"Invalid path {path}")
        return cls(
            mission=parts[0],
            level=ProcessingLevel(parts[1]),
            day=int(parts[4]),
            month=int(parts[3]),
            year=int(parts[2]),
            id=parts[5],
            path="/".join(parts[6:]),
        )



"""Handles operations related to STAC collections."""

from typing import Generator, Optional

from pystac import Collection, Extent, SpatialExtent, TemporalExtent
from pystac.utils import str_to_datetime

from datacosmos.datacosmos_client import DatacosmosClient
from datacosmos.exceptions.datacosmos_exception import DatacosmosException
from datacosmos.stac.collection.models.collection_update import CollectionUpdate
from datacosmos.utils.http_response.check_api_response import check_api_response


class CollectionClient:
    """Handles operations related to STAC collections."""

    def __init__(self, client: DatacosmosClient):
        """Initialize the CollectionClient with a DatacosmosClient."""
        self.client = client
        self.base_url = client.config.stac.as_domain_url()

    def fetch_collection(self, collection_id: str) -> Collection:
        """Fetch details of an existing STAC collection."""
        url = self.base_url.with_suffix(f"/collections/{collection_id}")
        response = self.client.get(url)
        check_api_response(response)
        return Collection.from_dict(response.json())

    def create_collection(self, collection: Collection) -> None:
        """Create a new STAC collection.

        Args:
            collection (Collection): The STAC collection to create.

        Raises:
            InvalidRequest: If the collection data is malformed.
        """
        if isinstance(collection.extent, dict):
            spatial_data = collection.extent.get("spatial", {}).get("bbox", [[]])
            temporal_data = collection.extent.get("temporal", {}).get("interval", [[]])

            # Convert string timestamps to datetime objects
            parsed_temporal = []
            for interval in temporal_data:
                start = str_to_datetime(interval[0]) if interval[0] else None
                end = (
                    str_to_datetime(interval[1])
                    if len(interval) > 1 and interval[1]
                    else None
                )
                parsed_temporal.append([start, end])

            collection.extent = Extent(
                spatial=SpatialExtent(spatial_data),
                temporal=TemporalExtent(parsed_temporal),
            )

        url = self.base_url.with_suffix("/collections")
        response = self.client.post(url, json=collection.to_dict())
        check_api_response(response)

    def update_collection(
        self, collection_id: str, update_data: CollectionUpdate
    ) -> None:
        """Update an existing STAC collection."""
        url = self.base_url.with_suffix(f"/collections/{collection_id}")
        response = self.client.patch(
            url, json=update_data.model_dump(by_alias=True, exclude_none=True)
        )
        check_api_response(response)

    def delete_collection(self, collection_id: str) -> None:
        """Delete a STAC collection by its ID."""
        url = self.base_url.with_suffix(f"/collections/{collection_id}")
        response = self.client.delete(url)
        check_api_response(response)

    def fetch_all_collections(self) -> Generator[Collection, None, None]:
        """Fetch all STAC collections with pagination support."""
        url = self.base_url.with_suffix("/collections")
        params = {"limit": 10}

        while True:
            data = self._fetch_collections_page(url, params)
            yield from self._parse_collections(data)

            next_cursor = self._get_next_pagination_cursor(data)
            if not next_cursor:
                break

            params["cursor"] = next_cursor

    def _fetch_collections_page(self, url: str, params: dict) -> dict:
        """Fetch a single page of collections from the API."""
        response = self.client.get(url, params=params)
        check_api_response(response)

        data = response.json()

        if isinstance(data, list):
            return {"collections": data}

        return data

    def _parse_collections(self, data: dict) -> Generator[Collection, None, None]:
        """Convert API response data to STAC Collection objects, ensuring required fields exist."""
        return (
            Collection.from_dict(
                {
                    **collection,
                    "type": collection.get("type", "Collection"),
                    "id": collection.get("id", ""),
                    "stac_version": collection.get("stac_version", "1.0.0"),
                    "extent": collection.get(
                        "extent",
                        {"spatial": {"bbox": []}, "temporal": {"interval": []}},
                    ),
                    "links": collection.get("links", []) or [],
                    "properties": collection.get("properties", {}),
                }
            )
            for collection in data.get("collections", [])
            if collection.get("type") == "Collection"
        )

    def _get_next_pagination_cursor(self, data: dict) -> Optional[str]:
        """Extract the next pagination token from the response."""
        next_href = self._get_next_link(data)
        return self._extract_pagination_token(next_href) if next_href else None

    def _get_next_link(self, data: dict) -> Optional[str]:
        """Extract the next page link from the response."""
        next_link = next(
            (link for link in data.get("links", []) if link.get("rel") == "next"), None
        )
        return next_link.get("href", "") if next_link else None

    def _extract_pagination_token(self, next_href: str) -> Optional[str]:
        """Extract the pagination token from the next link URL.

        Args:
            next_href (str): The next page URL.

        Returns:
            Optional[str]: The extracted token, or None if parsing fails.
        """
        try:
            return next_href.split("?")[1].split("=")[-1]
        except (IndexError, AttributeError) as e:
            raise DatacosmosException(
                f"Failed to parse pagination token from {next_href}",
                response=e.response,
            ) from e



"""STAC Client module for interacting with a STAC (SpatioTemporal Asset Catalog) API.

Provides methods for querying, fetching, creating, updating, and deleting STAC items.
"""

from typing import Generator, Optional

from pystac import Item

from datacosmos.datacosmos_client import DatacosmosClient
from datacosmos.exceptions.datacosmos_exception import DatacosmosException
from datacosmos.stac.item.models.catalog_search_parameters import (
    CatalogSearchParameters,
)
from datacosmos.stac.item.models.datacosmos_item import DatacosmosItem
from datacosmos.stac.item.models.item_update import ItemUpdate
from datacosmos.utils.http_response.check_api_response import check_api_response


class ItemClient:
    """Client for interacting with the STAC API."""

    def __init__(self, client: DatacosmosClient):
        """Initialize the STACClient with a DatacosmosClient.

        Args:
            client (DatacosmosClient): The authenticated Datacosmos client instance.
        """
        self.client = client
        self.base_url = client.config.stac.as_domain_url()

    def fetch_item(self, item_id: str, collection_id: str) -> Item:
        """Fetch a single STAC item by ID.

        Args:
            item_id (str): The ID of the item to fetch.
            collection_id (str): The ID of the collection containing the item.

        Returns:
            Item: The fetched STAC item.
        """
        url = self.base_url.with_suffix(f"/collections/{collection_id}/items/{item_id}")
        response = self.client.get(url)
        check_api_response(response)
        return Item.from_dict(response.json())

    def search_items(
        self, parameters: CatalogSearchParameters, project_id: str
    ) -> Generator[Item, None, None]:
        """Query the STAC catalog using the POST endpoint with filtering and pagination.

        Args:
            parameters (CatalogSearchParameters): The search parameters.

        Yields:
            Item: Parsed STAC item.
        """
        url = self.base_url.with_suffix("/search")
        parameters_query = parameters.to_query()
        body = {"project": project_id, "limit": 50, "query": parameters_query}
        if parameters.collections is not None:
            body = body | {"collections": parameters.collections}
        return self._paginate_items(url, body)

    def create_item(self, item: Item | DatacosmosItem) -> None:
        """Create a new STAC item in its own collection.

        The collection ID is inferred from the item.

        Args:
            item (Item | DatacosmosItem): The STAC item to be created.

        Raises:
            ValueError: If the item has no collection set.
            RequestError: If the API returns an error response.
        """
        if isinstance(item, Item):
            collection_id = item.collection_id or (
                item.get_collection().id if item.get_collection() else None
            )
        else:
            collection_id = item.collection

        if not collection_id:
            raise ValueError("Cannot create item: no collection_id found on item")

        url = self.base_url.with_suffix(f"/collections/{collection_id}/items")
        item_json: dict = item.to_dict()
        response = self.client.post(url, json=item_json)
        check_api_response(response)

    def add_item(self, item: Item | DatacosmosItem) -> None:
        """Adds item to catalog.

        The collection ID is inferred from the item.

        Args:
            item (Item | DatacosmosItem): The STAC item to be created.

        Raises:
            ValueError: If the item has no collection set.
            RequestError: If the API returns an error response.
        """
        if isinstance(item, Item):
            collection_id = item.collection_id or (
                item.get_collection().id if item.get_collection() else None
            )
        else:
            collection_id = item.collection

        if not collection_id:
            raise ValueError("Cannot create item: no collection_id found on item")

        url = self.base_url.with_suffix(f"/collections/{collection_id}/items/{item.id}")
        item_json: dict = item.to_dict()
        response = self.client.put(url, json=item_json)
        check_api_response(response)

    def update_item(
        self, item_id: str, collection_id: str, update_data: ItemUpdate
    ) -> None:
        """Partially update an existing STAC item.

        Args:
            item_id (str): The ID of the item to update.
            collection_id (str): The ID of the collection containing the item.
            update_data (ItemUpdate): The structured update payload.
        """
        url = self.base_url.with_suffix(f"/collections/{collection_id}/items/{item_id}")

        update_payload = update_data.model_dump(by_alias=True, exclude_none=True)

        if "assets" in update_payload:
            update_payload["assets"] = {
                key: asset.to_dict() for key, asset in update_payload["assets"].items()
            }
        if "links" in update_payload:
            update_payload["links"] = [
                link.to_dict() for link in update_payload["links"]
            ]

        response = self.client.patch(url, json=update_payload)
        check_api_response(response)

    def delete_item(self, item_id: str, collection_id: str) -> None:
        """Delete a STAC item by its ID.

        Args:
            item_id (str): The ID of the item to delete.
            collection_id (str): The ID of the collection containing the item.

        Raises:
            OCError: If the item is not found or deletion is forbidden.
        """
        url = self.base_url.with_suffix(f"/collections/{collection_id}/items/{item_id}")
        response = self.client.delete(url)
        check_api_response(response)

    def _paginate_items(self, url: str, body: dict) -> Generator[Item, None, None]:
        """Handle pagination for the STAC search POST endpoint.

        Fetches items one page at a time using the 'next' link.

        Args:
            url (str): The base URL for the search endpoint.
            body (dict): The request body containing search parameters.

        Yields:
            Item: Parsed STAC item.
        """
        params = {"limit": body.get("limit", 10)}

        while True:
            response = self.client.post(url, json=body, params=params)
            check_api_response(response)
            data = response.json()

            yield from (Item.from_dict(feature) for feature in data.get("features", []))

            next_href = self._get_next_link(data)
            if not next_href:
                break

            token = self._extract_pagination_token(next_href)
            if not token:
                break
            params["cursor"] = token

    def _get_next_link(self, data: dict) -> Optional[str]:
        """Extract the next page link from the response."""
        next_link = next(
            (link for link in data.get("links", []) if link.get("rel") == "next"), None
        )
        return next_link.get("href", "") if next_link else None

    def _extract_pagination_token(self, next_href: str) -> Optional[str]:
        """Extract the pagination token from the next link URL.

        Args:
            next_href (str): The next page URL.

        Returns:
            Optional[str]: The extracted token, or None if parsing fails.

        Raises:
            DatacosmosException: If pagination token extraction fails.
        """
        try:
            return next_href.split("?")[1].split("=")[-1]
        except (IndexError, AttributeError) as e:
            raise DatacosmosException(
                f"Failed to parse pagination token from {next_href}",
                response=e.response,
            ) from e



"""Configuration for local user account authentication.

When this is chosen, the user will be prompted to log in using their OPS credentials.
This will be used for running scripts locally.
"""

from typing import Literal

from pydantic import BaseModel


class LocalUserAccountAuthenticationConfig(BaseModel):
    """Configuration for local user account authentication.

    When this is chosen, the user will be prompted to log in using their OPS credentials.
    This will be used for running scripts locally.
    """

    type: Literal["local"]
    client_id: str
    authorization_endpoint: str
    token_endpoint: str
    redirect_port: int
    scopes: str
    audience: str
    cache_file: str



"""Authentication configuration options."""

from typing import Union

from datacosmos.config.models.local_user_account_authentication_config import (
    LocalUserAccountAuthenticationConfig,
)
from datacosmos.config.models.m2m_authentication_config import M2MAuthenticationConfig
from datacosmos.config.models.no_authentication_config import NoAuthenticationConfig

AuthenticationConfig = Union[
    NoAuthenticationConfig,
    LocalUserAccountAuthenticationConfig,
    M2MAuthenticationConfig,
]



"""Module for configuring machine-to-machine (M2M) authentication.

Used when running scripts in the cluster that require automated authentication
without user interaction.
"""

from typing import Literal

from pydantic import BaseModel, Field


class M2MAuthenticationConfig(BaseModel):
    """Configuration for machine-to-machine authentication.

    This is used when running scripts in the cluster that require authentication
    with client credentials.
    """

    DEFAULT_TYPE: Literal["m2m"] = "m2m"
    DEFAULT_TOKEN_URL: str = "https://login.open-cosmos.com/oauth/token"
    DEFAULT_AUDIENCE: str = "https://beeapp.open-cosmos.com"

    type: Literal["m2m"] = Field(default=DEFAULT_TYPE)
    client_id: str
    token_url: str = Field(default=DEFAULT_TOKEN_URL)
    audience: str = Field(default=DEFAULT_AUDIENCE)
    client_secret: str



"""URL utility class for building and handling URLs in the SDK."""


class URL:
    """Class to represent and build URLs in a convenient way."""

    def __init__(self, protocol: str, host: str, port: int, base: str):
        """Creates a new basis to build URLs.

        Args:
            protocol (str): Protocol to use in the URL (http/https).
            host (str): Hostname (e.g., example.com).
            port (int): Port number.
            base (str): Base path (e.g., /api/v1).
        """
        self.protocol = protocol
        self.host = host
        self.port = port
        self.base = base

    def string(self) -> str:
        """Returns the full URL as a string."""
        port = "" if self.port in [80, 443] else f":{self.port}"
        base = f"/{self.base.lstrip('/')}" if self.base else ""
        return f"{self.protocol}://{self.host}{port}{base}"

    def with_suffix(self, suffix: str) -> str:
        """Appends a suffix to the URL, ensuring proper formatting.

        Args:
            suffix (str): The path to append.

        Returns:
            str: Full URL with the suffix.
        """
        base = self.string()
        return f"{base.rstrip('/')}/{suffix.lstrip('/')}"

    def with_base(self, url: str) -> str:
        """Replaces the base of the url with the base stored in the URL object. (migrates url from one base to another).

        Args:
            url (str): url to migrate to the base of the URL object

        Returns (str):
            url with the base of the URL object
        """
        split_url = url.split("/")
        if len(split_url) < 3 or url.find("://") == -1:
            raise ValueError(f"URL '{url}' does not meet the minimum requirements")
        # get the whole path
        url_path = "/".join(split_url[3:])
        # simple case, matching self.base at url
        b = self.base.lstrip("/")
        if (base_pos := url_path.find(b)) != -1:
            # remove the base from the url
            url_suffix = url_path[len(b) + base_pos :]
        else:
            url_suffix = url_path
        return self.with_suffix(url_suffix)



"""Season enum class."""

from enum import Enum


class Season(str, Enum):
    """Different Open Cosmos seasons."""

    SUMMER = "Summer"
    WINTER = "Winter"
    AUTUMN = "Autumn"
    SPRING = "Spring"
    RAINY = "Rainy"
    DRY = "Dry"



"""Module defining a structured URL configuration model.

Ensures that URLs contain required components such as protocol, host,
port, and path.
"""

from pydantic import BaseModel

from datacosmos.utils.url import URL as DomainURL


class URL(BaseModel):
    """Generic configuration model for a URL.

    This class provides attributes to store URL components and a method
    to convert them into a `DomainURL` instance.
    """

    protocol: str
    host: str
    port: int
    path: str

    def as_domain_url(self) -> DomainURL:
        """Convert the URL instance to a `DomainURL` object.

        Returns:
            DomainURL: A domain-specific URL object.
        """
        return DomainURL(
            protocol=self.protocol,
            host=self.host,
            port=self.port,
            base=self.path,
        )
